import sys
import os
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))

import tensorflow_datasets as tfds
import pickle
import numpy as np
import tarfile
import gdown

import utils.utils as utils
from nkia.ml.cnn_nlp_model import CNN

class classifyProducts(object):

    def __init__(self):
        self.utils = utils

    def inference_from_model(self, ingredients=[''], allergics=[''], description=[''], product_name=['']): 
        """
        This method is responsible to infer the product category based on your caracteristics. Pass just one product
        by time.

        Args:
            ingredients (list): ['água açúcar baunilha']
            allergics (list): ['contém glúten, pode conter nozes']
            description (list):  ['Esse produto é simplesmente fantástico']
            product_name (list):  ['energético sabor baunilha']

        Returns:
            str: Category name. e.g: 'Bebidas não alcoolicas'

        """     
        if 'model_checkpoint' not in os.listdir():
            self.download_model(
                'https://drive.google.com/u/1/uc?export=download&confirm=pgRf&id=1jDL1j1cdEAPr4avB19bsx-90nnuCv1-s', 'nlp_model.tar.gz')

        X = self.utils.get_cleaned_predictor([ingredients], [allergics], [description], [product_name])
        X = self.preprocessing_fn(X)

        label_encoder = self.get_label_encoder()

        cnn = CNN(vocab_size=self.tokenizer.vocab_size, emb_dim=300, nb_filters=100,
            ffn_units=256, nb_classes=29, dropout_rate=0.2)

        cnn.load_weights('./model_checkpoint/cp.ckpt')

        product_category = label_encoder.inverse_transform(
            [cnn(np.array(X), training=False).numpy().argmax()])[0]

        return product_category

    def download_model(self, url, output):
        gdown.download(url, output, quiet=False) 

        tar = tarfile.open(output)
        tar.extractall()
        tar.close()

    def preprocessing_fn(self, X):
        self.tokenizer = tfds.features.text.SubwordTextEncoder.load_from_file('token')
        tokenized_predictor = [self.tokenizer.encode(sentence) for sentence in  X] 

        return tokenized_predictor

    def get_label_encoder(self):
        pkl_encoder_file = open('label_encoder.pkl', 'rb')
        encoder = pickle.load(pkl_encoder_file) 
        pkl_encoder_file.close()
        return encoder

classifyProducts()


import pandas as pd
import pymongo
import unidecode
import re 
import string
import nltk

def get_dataframe_from_mongo(mongo_conn):
    products_df = mongo_conn.db.df_products.find({}).sort('created', pymongo.DESCENDING)
    
    if products_df:
        return pd.DataFrame(products_df[0]['df'])

    return ''

def get_cleaned_predictor(ingredients, allergics, description, product_name):
    ingredients_and_allergics = list(map(get_ingredients_and_allergics, ingredients, allergics))
    description = list(map(get_description_and_name, description))
    product_name = list(map(get_description_and_name, product_name))

    X = join_all_predictors(ingredients_and_allergics, description, product_name)

    return X

def get_ingredients_and_allergics(ingredients, allergics):
    ingredients += allergics if allergics != '[]' else ''   
    ingredient = list(map(lambda x: unidecode.unidecode(x).lower(), ingredients))
    
    return ' '.join(ingredient)

def get_description_and_name(data):    
    stop_words = nltk.corpus.stopwords.words('portuguese')

    data = unidecode.unidecode(data[0]).lower()
    data = re.sub(r'['+string.punctuation+']', '', data)
    data = [plv for plv in data.split() if plv not in stop_words]
    
    return ' '.join(data)   

def join_all_predictors(ingredients_and_allergics, description, product_name):
    count = 0
    X = []

    for name in product_name:
        X.append(name + ' ' + description[count] + ' ' + ingredients_and_allergics[count])
        count += 1

    return X


import sys
import os
from sys import argv
sys.path.insert(0, os.path.join(os.path.dirname(__file__), '../../'))

import tensorflow_datasets as tfds
import tensorflow as tf
from sklearn.preprocessing import LabelEncoder
from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix

import matplotlib.pyplot as plt
import seaborn as sns
import pandas as pd
import pickle

import utils.utils as utils
from database.db_mongo import MongoDatabase
from nkia.ml.cnn_nlp_model import CNN

class trainProductsClassifier(object):

    def __init__(self, environment):
        self.utils = utils
        self.environment = environment
        self.train_model()

    def train_model(self):
        dataframe, y = self.get_data()
        X = self.get_previsors_attrib(dataframe)
        
        X_train, X_test, y_train, y_test = train_test_split(X,
                                                            y,
                                                            test_size=0.3) 

        cnn = self.create_model_instance(y)
        cnn.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        cp_callback = tf.keras.callbacks.ModelCheckpoint(filepath='./src/nkia/ml/saved_model/nlp_model/model_checkpoint/cp.ckpt',
                                                        save_weights_only=True,
                                                        verbose=1)

        cnn.fit(X_train, y_train,
                batch_size = 64,
                epochs = 1,
                verbose = 1,
                validation_split = 0.10,
                callbacks=[cp_callback])

        self.evaluate_model(cnn, X_test, y_test)

    def get_data(self):
        self.open_mongo_connections()
        dataframe = self.utils.get_dataframe_from_mongo(self.mongo_conn)
        self.close_mongo_connection()

        dataframe['allergics'].fillna('[]', inplace=True)
        dataframe['description'].fillna('', inplace=True)

        y = self.encode_class_value(dataframe['Category'].values)

        print('Restored data from Mongo')
        sys.stdout.flush()
        return dataframe, y

    def open_mongo_connections(self):
        self.mongo = MongoDatabase(self.environment)
        self.mongo.connect()
        self.mongo_conn = self.mongo

    def close_mongo_connection(self):
        self.mongo.close_connection()

    def encode_class_value(self, class_value):
        encoder = LabelEncoder()
        y = encoder.fit_transform(class_value)

        output_encoder = open('./src/nkia/ml/saved_model/nlp_model/label_encoder.pkl', 'wb')
        pickle.dump(encoder, output_encoder)
        output_encoder.close()
        
        return y

    def get_previsors_attrib(self, dataframe):
        ingredients = [x for x in dataframe['ingredients']] 
        allergics = [x for x in dataframe['allergics']]
        description = [[x] for x in dataframe['description']]
        product_name = [[x] for x in dataframe['name']]

        X = self.utils.get_cleaned_predictor(ingredients, allergics, description, product_name)
        
        X = self.preprocessing_fn(X)

        return X

    def preprocessing_fn(self, X):
        X = self.tokenize_data(X)
        X = self.padding_matrix(X)

        print('Preprocessing_fn finished')
        sys.stdout.flush()
        return X

    def tokenize_data(self, data):
        self.tokenizer = tfds.features.text.SubwordTextEncoder.build_from_corpus(
            data, target_vocab_size=2**14)

        self.tokenizer.save_to_file('./src/nkia/ml/saved_model/nlp_model/token')
        data = [self.tokenizer.encode(sentence) for sentence in data]

        return data

    def padding_matrix(self, data):
        max_sentence_len = max([len(sentence) for sentence in data])
        data = tf.keras.preprocessing.sequence.pad_sequences(data,
                                                             value=0,
                                                             padding='post',
                                                             maxlen=max_sentence_len)

        return data

    def create_model_instance(self, y):
        vocab_size = self.tokenizer.vocab_size
        nb_classes = len(set(y))

        cnn = CNN(vocab_size=vocab_size, emb_dim=300, nb_filters=100,
            ffn_units=256, nb_classes=nb_classes, dropout_rate=0.2)

        print('Created model instance')
        sys.stdout.flush()
        return cnn
    
    def evaluate_model(self, cnn, X_test, y_test, vz_results=False):
        results = cnn.evaluate(X_test, y_test, batch_size=64)
        print('Loss and Accuracy: ', results)

        if vz_results:
            y_pred_test = cnn.predict(X_test)
            predicted = [result.argmax() for result in y_pred_test]
            fig, ax = plt.subplots(figsize=(10,10))
            cm = confusion_matrix(y_test, predicted)
            sns.heatmap(cm, annot=True, vmin=0, vmax=1)
            plt.show()


trainProductsClassifier(argv[1] if len(argv) > 1 else 'dev')
